@@ -4,9 +4,6 @@
 #include <gdk-pixbuf/gdk-pixbuf.h>
 #include <gnome.h>
 #include "minefield.h"
-#include "flag.xpm"
-#include "mine.xpm"
-#include "question.xpm"
 
 static struct {
 	gint x;
@@ -52,7 +49,6 @@ static gint minefield_signals[LAST_SIGNAL] = { 0 };
 static GtkWidgetClass *parent_class;
 
 /*  Prototypes */
-static gulong get_random(gulong limit);
 static inline gint cell_idx(GtkMineField *mfield, guint x, guint y);
 static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y);
 static gint gtk_minefield_button_press(GtkWidget *widget, GdkEventButton *event);
@@ -66,16 +62,15 @@ static gint gtk_minefield_motion_notify(GtkWidget *widget, GdkEventMotion *event
 static void gtk_minefield_multi_release (GtkMineField *mfield, guint x, guint y, guint c, guint really);
 static void gtk_minefield_randomize(GtkMineField *mfield, int curloc);
 static void gtk_minefield_realize(GtkWidget *widget);
-static void gtk_minefield_setup_signs(GtkWidget *widget);
+static void gtk_minefield_setup_signs(GtkMineField *mfield);
 static void gtk_minefield_show(GtkMineField *mfield, guint x, guint y);
 static void gtk_minefield_size_allocate(GtkWidget *widget, GtkAllocation *allocation);
 static void gtk_minefield_size_request(GtkWidget *widget, GtkRequisition *requisition);
 static void gtk_minefield_toggle_mark(GtkMineField *mfield, guint x, guint y);
 static void gtk_minefield_unrealize (GtkWidget *widget);
 static void gtk_minefield_win(GtkMineField *mfield);
-static void init_random(gulong seed);
 static inline int gtk_minefield_check_cell(GtkMineField *mfield, guint x, guint y);
-static void _setup_sign (sign *signp, const char **data, guint minesize);
+static void _setup_sign (sign *signp, const char *file, guint minesize);
 static inline void gtk_minefield_multi_press(GtkMineField *mfield, guint x, guint y, gint c);
 /* end prototypes */
 
@@ -88,40 +83,48 @@ static inline gint cell_idx(GtkMineField *mfield, guint x, guint y)
 	return -1;
 }
 
-static void _setup_sign (sign *signp, const char **data, guint minesize)
+static void _setup_sign (sign *signp, const char *file, guint minesize)
 {
-	GdkPixbuf *image;
+	GdkPixbuf *scaledimage;
 
-	image = gdk_pixbuf_new_from_xpm_data ( (const char **) data);
+	if (signp->pixbuf == NULL) {
+		GError *error = NULL;
+		signp->pixbuf = gdk_pixbuf_new_from_file (file, &error);
+		if (signp->pixbuf == NULL)
+			g_error (error->message);
+	}	
 
-	gdk_pixbuf_render_pixmap_and_mask (image, &signp->pixmap,
+        scaledimage = gdk_pixbuf_scale_simple (signp->pixbuf, minesize - 2, minesize - 2,
+                                               GDK_INTERP_BILINEAR);
+        
+	gdk_pixbuf_render_pixmap_and_mask (scaledimage, &signp->pixmap,
 					   &signp->mask, 127);
 
-	gdk_pixbuf_unref (image);
-        gdk_drawable_get_size(signp->pixmap, &(signp->width), &(signp->height));
+        g_object_unref (scaledimage);
+	gdk_drawable_get_size(signp->pixmap, &(signp->width), &(signp->height));
 }
 
-static void gtk_minefield_setup_signs(GtkWidget *widget)
+static void gtk_minefield_setup_signs(GtkMineField *mfield)
 {
-        GtkMineField *mfield;
-  
-        mfield = GTK_MINEFIELD(widget);
-  
-        _setup_sign(&mfield->flag, flag_xpm, mfield->minesize);
-        _setup_sign(&mfield->mine, mine_xpm, mfield->minesize);
-	_setup_sign(&mfield->question, question_xpm, mfield->minesize);
+        _setup_sign(&mfield->flag, DATADIR"/pixmaps/gnomine/flag.png", mfield->minesize);
+        _setup_sign(&mfield->mine, DATADIR"/pixmaps/gnomine/mine.png", mfield->minesize);
+        _setup_sign(&mfield->question, DATADIR"/pixmaps/gnomine/flag-question.png", mfield->minesize);
 }
 
 static void
-gtk_minefield_setup_numbers(GtkMineField *mfield)
+gtk_minefield_setup_numbers (GtkMineField *mfield)
 {
 	int minesize, pixel_sz, i;
+	static guint last_size = 0;
 	  
 	minesize = mfield->minesize;
 
 	pixel_sz = minesize - 2;
 	if (pixel_sz > 999) pixel_sz = 999;
-	if (pixel_sz < 2)  pixel_sz = 2;
+
+	if (last_size == pixel_sz)
+		return;
+	last_size = pixel_sz;
   
 	for (i=0; i<9; i++) {
 		gchar text[2];
@@ -211,50 +214,51 @@ static void gtk_minefield_realize(GtkWidget *widget)
         
         widget->style = gtk_style_attach(widget->style, widget->window);
         gtk_style_set_background(widget->style, widget->window, GTK_STATE_ACTIVE);
-
-        gtk_minefield_setup_signs(widget);
 }
 
 static void gtk_minefield_unrealize (GtkWidget *widget)
 {
-	GtkMineField *mfield;
-
 	g_return_if_fail (widget != NULL);
 	g_return_if_fail (GTK_IS_MINEFIELD (widget));
 
-	mfield = GTK_MINEFIELD (widget);
-
 	if (GTK_WIDGET_CLASS (parent_class)->unrealize)
 		(* GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
 }
 
-static void gtk_minefield_size_allocate(GtkWidget *widget,
-					GtkAllocation *allocation)
+/* The frame makes sure that the minefield is allocated the correct size */
+/* This is the standard allocate routine - it could be removed and the parents routine inherited */
+static void gtk_minefield_size_allocate (GtkWidget     *widget,
+				 	 GtkAllocation *allocation)
 {
-        g_return_if_fail(widget != NULL);
-        g_return_if_fail(GTK_IS_MINEFIELD (widget));
-        g_return_if_fail(allocation != NULL);
+        guint minesize, width, height;
+        guint xofs, yofs;
+        GtkMineField * mfield;
 
         widget->allocation = *allocation;
+
+        mfield = GTK_MINEFIELD (widget);
         
-	if (GTK_WIDGET_REALIZED(widget)) {
-		gdk_window_move_resize(widget->window,
-				       allocation->x, allocation->y,
-				       GTK_MINEFIELD(widget)->xsize *
-				       GTK_MINEFIELD(widget)->minesize,
-				       GTK_MINEFIELD(widget)->ysize *
-				       GTK_MINEFIELD(widget)->minesize);
+	if (GTK_WIDGET_REALIZED (widget)) {
+                minesize = MIN (allocation->width / mfield->xsize,
+                            allocation->height / mfield->ysize);
+                mfield->minesize = minesize;
+                width = mfield->xsize * minesize;
+                height = mfield->ysize * minesize;
+                xofs = allocation->x + (allocation->width - width)/2;
+                yofs = allocation->y + (allocation->height - height)/2;
+                
+		gdk_window_move_resize (widget->window, xofs, yofs,
+                                        width, height);
 	}
 }
 
-static void gtk_minefield_size_request(GtkWidget *widget,
-				       GtkRequisition *requisition)
+static void gtk_minefield_size_request (GtkWidget      *widget,
+				        GtkRequisition *requisition)
 {
-        requisition->width  = GTK_MINEFIELD(widget)->xsize *
-		              GTK_MINEFIELD(widget)->minesize;
-
-	requisition->height = GTK_MINEFIELD(widget)->ysize * 
-		              GTK_MINEFIELD(widget)->minesize;
+        GtkMineField *mf = GTK_MINEFIELD (widget);
+	/* request the minimum size - to allow the widget window to be resized */
+	requisition->width  = mf->xsize * MINESIZE_MIN;
+	requisition->height = mf->ysize * MINESIZE_MIN;
 }
 
 static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
@@ -263,9 +267,22 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
 	int noshadow;
 	int n;
 	guint minesize;
-
+        static GdkGC *dots;
+        static char stipple_data[]  = { 0x03, 0x03, 0x0c, 0x0c };
+        static GdkPixmap *stipple = NULL;
         GtkWidget *widget = GTK_WIDGET(mfield);
 
+        /* This gives us a dotted line to increase the contrast between
+         * buttons and the "sea". */
+        if (stipple == NULL) {
+          stipple = gdk_bitmap_create_from_data (NULL, stipple_data, 4, 4);
+          dots = gdk_gc_new (widget->window);
+          gdk_gc_copy (dots, widget->style->dark_gc[2]);
+          gdk_gc_set_stipple (dots, stipple);
+          g_object_unref (stipple);
+          gdk_gc_set_fill (dots, GDK_STIPPLED);
+        }
+                
 	minesize = mfield->minesize;
 
 	if (mfield->lose || mfield->win) {
@@ -280,7 +297,7 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
 	if (noshadow) { /* draw grid on ocean floor */
 		if (y == 0) {	/* top row only */
 			gdk_draw_line(widget->window,	/* top */
-			              widget->style->dark_gc[2],
+			              dots,
 	                              x*minesize,
 	                              0,
 	                              x*minesize+minesize-1,
@@ -288,20 +305,20 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
 		}
 		if (x == 0) {	/* left column only */
 			gdk_draw_line(widget->window,	/* left */
-			              widget->style->dark_gc[2],
+			              dots,
 	                              0,
 	                              y*minesize,
 	                              0,
 	                              y*minesize+minesize-1);
 		}
 		gdk_draw_line(widget->window,	/* right */
-		              widget->style->dark_gc[2],
+                              dots,
                               x*minesize+minesize-1,
                               y*minesize,
                               x*minesize+minesize-1,
                               y*minesize+minesize-1);
 		gdk_draw_line(widget->window,	/* bottom */
-		              widget->style->dark_gc[2],
+		              dots,
                               x*minesize,
                               y*minesize+minesize-1,
                               x*minesize+minesize-1,
@@ -401,48 +418,44 @@ static void gtk_mine_draw(GtkMineField *mfield, guint x, guint y)
 	}
 }
 
-void gtk_minefield_draw(GtkMineField *mfield, GdkRectangle *area)
+static gint gtk_minefield_expose (GtkWidget      *widget,
+				  GdkEventExpose *event)
 {
-	guint x1, y1, x2, y2, x, y, minesize;
+	g_return_val_if_fail (widget != NULL, FALSE);
+        g_return_val_if_fail (GTK_IS_MINEFIELD (widget), FALSE);
+        g_return_val_if_fail (event != NULL, FALSE);
 
-	minesize = mfield->minesize;
+	if (GTK_WIDGET_DRAWABLE (widget)) {
+		guint x1, y1, x2, y2, x, y;
+		GtkMineField *mfield = GTK_MINEFIELD (widget);
+		GdkRectangle *area = &event->area;
 
-	if (area) {
-		x1 = area->x/minesize;
-		y1 = area->y/minesize;
-		x2 = (area->x + area->width - 1) / minesize;
-		y2 = (area->y + area->height - 1) / minesize;
-	} else {
-		x1 = 0; y1 = 0;
-		x2 = mfield->xsize;
-		y2 = mfield->ysize;
+		/* mine square numbers must be resized to fit the mine size */
+                gtk_minefield_setup_signs (mfield);
+		gtk_minefield_setup_numbers (mfield);
+
+		if (area) {
+			x1 = area->x/mfield->minesize;
+			y1 = area->y/mfield->minesize;
+			x2 = (area->x + area->width - 1) / mfield->minesize;
+			y2 = (area->y + area->height - 1) / mfield->minesize;
+		} else {
+			x1 = 0; y1 = 0;
+			x2 = mfield->xsize - 1;
+			y2 = mfield->ysize - 1;
+		}
+
+		/* These are necessary because we get an expose call before a
+		 * resize at the old size, but after we have changed our data. */
+		if (x2 >= mfield->xsize)
+			x2 = mfield->xsize - 1;
+		if (y2 >= mfield->ysize)
+			y2 = mfield->ysize - 1;
+		
+		for (x = x1; x <= x2; x++)
+			for (y = y1; y <= y2; y++)
+				gtk_mine_draw (mfield, x, y);
 	}
-	
-	for (x = x1; x <= x2; x++)
-		for (y = y1; y <= y2; y++)
-			gtk_mine_draw(mfield, x, y);
-}
-
-static gint gtk_minefield_expose(GtkWidget *widget,
-				 GdkEventExpose *event)
-{
-        GtkMineField *mfield;
-	guint minesize;
-
-	g_return_val_if_fail(widget != NULL, FALSE);
-        g_return_val_if_fail(GTK_IS_MINEFIELD(widget), FALSE);
-        g_return_val_if_fail(event != NULL, FALSE);
-
-	if (!GTK_WIDGET_DRAWABLE(widget))
-                return FALSE;
-
-        mfield = GTK_MINEFIELD(widget);
-	
-	minesize = mfield->minesize;
-
-
-	gtk_minefield_draw(GTK_MINEFIELD(widget), &event->area);
-
 	return FALSE;
 }
 
@@ -478,13 +491,11 @@ static void gtk_minefield_check_field(GtkMineField *mfield, gint x, gint y)
         guint changed;
 
         gint x1, y1, x2, y2;
-	gint cx1,cy1,cx2,cy2;
-	
-	cx1 = x-2;
-	cy1 = y-2;
-	cx2 = x+2;
-	cy2 = y+2;
+        gint cx1, cx2, cy1, cy2;
 
+        cx1 = cx2 = x;
+        cy1 = cy2 = y;
+        
         do {
 		x1 = cx1-1;
 		y1 = cy1-1;
@@ -515,13 +526,13 @@ static void gtk_minefield_check_field(GtkMineField *mfield, gint x, gint y)
         } while (changed);
 }
 
-static void gtk_minefield_lose(GtkMineField *mfield)
+static void gtk_minefield_lose (GtkMineField *mfield)
 {
-        g_signal_emit(GTK_OBJECT(mfield),
+        g_signal_emit (G_OBJECT (mfield),
                         minefield_signals[EXPLODE_SIGNAL],
 			0, NULL );
         mfield->lose = 1;
-        gtk_minefield_draw(mfield, NULL);
+	gtk_widget_queue_draw (GTK_WIDGET (mfield));
 }
 
 static void gtk_minefield_win(GtkMineField *mfield)
@@ -553,28 +564,24 @@ static void gtk_minefield_win(GtkMineField *mfield)
         mfield->win = 1;
 }
 
-static void gtk_minefield_randomize(GtkMineField *mfield, int curloc)
+static void gtk_minefield_randomize (GtkMineField *mfield, int curloc)
 {
-	guint i, j;
+	guint i;
 	guint x, y;
-	guint tmp;
 	guint n;
 	guint cidx;
+	static GRand *grand = NULL;
 
-	/* make sure curloc is not mined */
-	mfield->mines[mfield->xsize * mfield->ysize -1].mined =
-		mfield->mines[curloc].mined; /* last location is never mined */
-	mfield->mines[curloc].mined = 0;
+	if (grand == NULL)
+		grand = g_rand_new ();
 
-	/* swap mines randomly */
-	for (i=0; i<mfield->xsize * mfield->ysize; i++) {
-		j = (guint)get_random(mfield->xsize * mfield->ysize);
-		if (i == curloc || j == curloc) { /* skip current location */
-			continue;
+	/* randomly set the mines, but avoid the current location (why ?)*/
+	for (n = 0; n < mfield->mcount; ) {
+	        i = g_rand_int_range (grand, 0, mfield->xsize * mfield->ysize);
+		if (!mfield->mines[i].mined && i != curloc) {
+		        mfield->mines[i].mined = 1;
+			n++;
 		}
-		tmp = mfield->mines[i].mined;
-		mfield->mines[i].mined = mfield->mines[j].mined;
-		mfield->mines[j].mined = tmp;
 	}
 
 	/* load neighborhood numbers */
@@ -633,7 +640,11 @@ static void gtk_minefield_toggle_mark(GtkMineField *mfield, guint x, guint y)
 		mfield->flag_count++;
 		break;
 	case MINE_MARKED:
-		mfield->mines[c].marked = MINE_QUESTION;
+		if (mfield->use_question_marks) {
+			mfield->mines[c].marked = MINE_QUESTION;
+		} else {
+			mfield->mines[c].marked = MINE_NOMARK;
+		}
 		mfield->flag_count--;
 		break;
 	case MINE_QUESTION:
@@ -860,11 +871,8 @@ static gint gtk_minefield_button_release(GtkWidget *widget, GdkEventButton *even
 
 static void gtk_minefield_class_init (GtkMineFieldClass *class)
 {
-	GtkWidgetClass *widget_class;
-	GtkObjectClass *object_class;
-	
-        widget_class = (GtkWidgetClass *)class;
-	object_class = (GtkObjectClass *)class;
+ 	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
+ 	GtkObjectClass *object_class = GTK_OBJECT_CLASS (class);
 
 	parent_class = gtk_type_class (gtk_widget_get_type ());
 	
@@ -935,28 +943,13 @@ static void gtk_minefield_init (GtkMineField *mfield)
 {
         mfield->xsize = 0;
         mfield->ysize = 0;
+	mfield->mines = NULL;
+	mfield->started = FALSE;
+	mfield->cdown = -1;
 
-        GTK_WIDGET (mfield)->requisition.width = mfield->minesize;
-        GTK_WIDGET (mfield)->requisition.height = mfield->minesize;
-
-	gtk_minefield_setup_signs(GTK_WIDGET(mfield));
-
-	gtk_minefield_setup_numbers(mfield);
-}
-
-static gulong random_seed;
-
-static void init_random(gulong seed)
-{
-	random_seed = seed;
-}
-
-static gulong get_random(gulong limit)
-{
-	do {
-		random_seed = (random_seed*1139113+10921)>>2;
-	} while (random_seed > ((gulong)(G_MAXLONG/limit))*limit);
-	return random_seed % limit;
+	mfield->flag.pixbuf = NULL;
+        mfield->mine.pixbuf = NULL;
+	mfield->question.pixbuf = NULL;
 }
 
 void gtk_minefield_set_size(GtkMineField *mfield, guint xsize, guint ysize)
@@ -973,31 +966,22 @@ void gtk_minefield_set_size(GtkMineField *mfield, guint xsize, guint ysize)
 		mfield->xsize = xsize;
 		mfield->ysize = ysize;
 		if (GTK_WIDGET_VISIBLE(mfield)) {
-			gtk_widget_queue_resize(GTK_WIDGET(mfield));
+			gtk_widget_queue_resize(GTK_WIDGET(mfield)); 
 		}
 	}
 }
 
-GtkWidget* gtk_minefield_new(void)
+GtkWidget* gtk_minefield_new (void)
 {
-        GtkMineField *mfield;
-        
-	mfield = GTK_MINEFIELD (g_object_new (TYPE_GTK_MINEFIELD,
-				NULL));
-	mfield->mines = NULL;
-	mfield->started = FALSE;
-	gtk_minefield_set_size(mfield, 0, 0);
-
-	mfield->cdown = -1;
-	return GTK_WIDGET(mfield);
+	return GTK_WIDGET (g_object_new (GTK_TYPE_MINEFIELD, NULL));
 }
 
-guint gtk_minefield_get_type ()
+GType gtk_minefield_get_type (void)
 {
         static GType minefield_type = 0;
         
 	if (minefield_type == 0) {
-                static GTypeInfo minefield_info =
+                static const GTypeInfo minefield_info =
                         {
                                 sizeof (GtkMineFieldClass),
 				NULL, /* base_init */
@@ -1019,29 +1003,12 @@ guint gtk_minefield_get_type ()
         return minefield_type;
 }
 
-void gtk_minefield_set_mines(GtkMineField *mfield, guint mcount, guint minesize)
-{
-        g_return_if_fail(mfield != NULL);
-        g_return_if_fail(GTK_IS_MINEFIELD(mfield));
-	g_return_if_fail(minesize>0);
-	
-        mfield->mcount = mcount;
-
-	if (mfield->minesize != minesize) {
-		mfield->minesize = minesize;
-		gtk_minefield_setup_numbers(mfield);
-		if (GTK_WIDGET_VISIBLE(mfield)) {
-			gtk_widget_queue_resize(GTK_WIDGET(mfield));
-		}
-	}
-}
-
-void gtk_minefield_restart(GtkMineField *mfield)
+void gtk_minefield_restart (GtkMineField *mfield)
 {
 	guint i;
 
-        g_return_if_fail(mfield != NULL);
-        g_return_if_fail(GTK_IS_MINEFIELD(mfield));
+        g_return_if_fail (mfield != NULL);
+        g_return_if_fail (GTK_IS_MINEFIELD (mfield));
 
 	mfield->flag_count = 0;
 	mfield->shown = 0;
@@ -1054,21 +1021,23 @@ void gtk_minefield_restart(GtkMineField *mfield)
         mfield->multi_mode = 0;
 	mfield->in_play = 0;
 
-	/* set up mfield->mcount mines.  (randomize later) */
-	for (i=0; i<mfield->xsize * mfield->ysize; i++) {
-		mfield->mines[i].mined = i<mfield->mcount ? 1 : 0;
+	for (i=0; i < mfield->xsize * mfield->ysize; i++) {
 		mfield->mines[i].marked = MINE_NOMARK;
+		mfield->mines[i].mined  = 0;
                 mfield->mines[i].shown  = 0;
 		mfield->mines[i].down   = 0;
 	}
 
-	if (secs == 0) {	/* random not seeded yet */
-		time((time_t *)&secs);
-		init_random(secs);
-	}
-	if (mfield->started == FALSE)
+        if (mfield->started == FALSE)
 		mfield->started = TRUE;
 	else
-		gtk_minefield_draw(mfield, NULL);
+                gtk_widget_queue_draw (GTK_WIDGET (mfield));
 }
 
+void gtk_minefield_set_use_question_marks(GtkMineField *mfield, gboolean use_question_marks)
+{
+	g_return_if_fail(mfield != NULL);
+	g_return_if_fail(GTK_IS_MINEFIELD(mfield));
+	
+	mfield->use_question_marks = use_question_marks;
+}
